from pathlib import Path


class S3AwarePath(type(Path())):
    def is_s3(self):
        return self.parts[0] == "s3:"

    def __str__(self):
        if self.is_s3():
            return f"s3://{super().__str__()[4:]}"
        else:
            return super().__str__()

    def get_s3_bucket(self):
        if self.is_s3():
            return self.parts[1]
        else:
            raise Exception("Not an S3 path")

    def get_s3_prefix(self):
        if self.is_s3():
            return "/".join(self.parts[2:])
        else:
            raise Exception("Not an S3 path")

    def get_table_name(self):
        if self.stem.count("_") < 1:
            raise Exception(
                "Not a valid format. Needs at least 1 `_` to match format `<schema_name>_<table_name>`"
            )
        schema_name = self.stem.split("_")[0]
        table_name = self.stem[len(schema_name) + 1 :]
        return schema_name, table_name



from pytest import raises


from ezduckdb import DuckDB, SQL
import pandas as pd


class TestSQL:
    def test_normal_bindings(self):
        inp = SQL("SELECT * FROM $table WHERE id = $id", table="foo", id=1)
        assert inp.to_string() == "SELECT * FROM 'foo' WHERE id = 1"

    def test_nonexistant_bindings(self):
        with raises(AssertionError) as _:
            SQL("SELECT * FROM table", id=1)

    def test_df_bindings(self):
        import pandas as pd

        df = pd.DataFrame({"id": [1, 2, 3]})
        inp = SQL("SELECT * FROM $table", table=df)
        assert inp.to_string() == "SELECT * FROM df_" + str(id(df))

    def test_sql_bindings(self):
        inp = SQL("SELECT * FROM $table", table=SQL("SELECT * FROM foo"))
        assert inp.to_string() == "SELECT * FROM (SELECT * FROM foo)"

    def test_recursive_sql_bindings(self):
        inp = SQL(
            "SELECT * FROM $table", table=SQL("SELECT * FROM $table", table="foo")
        )
        assert inp.to_string() == "SELECT * FROM (SELECT * FROM 'foo')"


class TestDuckDB:
    def test_query(self):
        db = DuckDB(s3_storage_used=False)
        assert db.query(SQL("select 1")).values == pd.DataFrame([(1,)]).values

    def test_context_manager(self):
        with DuckDB(s3_storage_used=False) as conn:
            assert conn.query("select 1").df().values == pd.DataFrame([(1,)]).values

    def test_df_bindings(self):
        df = pd.DataFrame({"id": [1, 2, 3]})
        db = DuckDB(s3_storage_used=False)
        act = db.query(SQL("SELECT * FROM $table", table=df)).values
        exp = pd.DataFrame([(1,), (2,), (3,)]).values
        assert (act == exp).all()

    def test_sql_bindings(self):
        db = DuckDB(s3_storage_used=False)
        act = db.query(
            SQL("SELECT * FROM $table", table=SQL("SELECT $val", val=1))
        ).values
        exp = pd.DataFrame([(1,)]).values
        assert (act == exp).all()

    def test_query_s3_parquet(self):
        s3_parquet = (
            "s3://codenym-automated-testing/ezduckdb/parquet/schema1_table1.parquet"
        )
        db = DuckDB(s3_storage_used=True)
        act = db.query(
            SQL("SELECT * FROM read_parquet($s3_parquet)", s3_parquet=s3_parquet)
        ).values
        exp = pd.DataFrame([[1, 4], [2, 5], [3, 6]]).values

        assert (act == exp).all()



from ezduckdb import S3AwarePath
import pytest


class TestS3AwarePath:
    def test_one(self):
        inp = S3AwarePath("s3://bucket/curated/s1chema_table1.csv")
        assert inp.get_s3_bucket() == "bucket"
        assert inp.get_s3_prefix() == "curated/s1chema_table1.csv"
        assert str(inp) == "s3://bucket/curated/s1chema_table1.csv"
        assert inp.is_s3() == True

    def test_two(self):
        inp = S3AwarePath("curated/s1chema_table1.csv")
        with pytest.raises(Exception) as _:
            inp.get_s3_bucket()
        with pytest.raises(Exception) as _:
            assert inp.get_s3_prefix()
        assert str(inp) == "curated/s1chema_table1.csv"
        assert inp.is_s3() == False


class Test_get_table_name_from_path:
    def test_one(self):
        inp = S3AwarePath("curated/s1chema_table1.csv")
        assert get_table_name_from_path(inp) == ("s1chema", "table1")

    def test_two(self):
        inp = S3AwarePath("curated/s1chema_table1_blah.parquet")
        assert get_table_name_from_path(inp) == ("s1chema", "table1_blah")

    def test_three(self):
        inp = S3AwarePath("s3://bucket/curated/s1chema_table2_blah.parquet")
        assert get_table_name_from_path(inp) == ("s1chema", "table2_blah")



from duckdb import connect
import pandas as pd
from sqlescapy import sqlescape
from string import Template
from typing import Mapping
from datanym import S3AwarePath


class SQL:
    def __init__(self, sql, **bindings):
        for binding in bindings:
            assert binding in sql
        self.sql = sql
        self.bindings = bindings

    def to_string(self) -> str:
        replacements = {}
        for key, value in self.bindings.items():
            if isinstance(value, pd.DataFrame):
                replacements[key] = f"df_{id(value)}"
            elif isinstance(value, SQL):
                replacements[key] = f"({value.to_string()})"
            elif isinstance(value, (str, S3AwarePath)):
                replacements[key] = f"'{sqlescape(value)}'"
            elif isinstance(value, (int, float, bool)):
                replacements[key] = str(value)
            elif value is None:
                replacements[key] = "null"
            else:
                raise ValueError(f"Invalid type for {key}")
        return Template(self.sql).safe_substitute(replacements)

    def collect_dataframes(self) -> Mapping[str, pd.DataFrame]:
        dataframes = {}
        for key, value in self.bindings.items():
            if isinstance(value, pd.DataFrame):
                dataframes[f"df_{id(value)}"] = value
            elif isinstance(value, SQL):
                dataframes.update(value.collect_dataframes())
        return dataframes


class DuckDB:
    def __init__(
        self,
        options="",
        db_location=":memory:",
        s3_storage_used=True,
        aws_profile="codenym",
    ):
        self.options = options
        self.db_location = db_location
        self.s3_storage_used = s3_storage_used
        self.aws_profile = aws_profile

    def connect(self):
        db = connect(self.db_location)
        if self.s3_storage_used:
            db.query("install httpfs; load httpfs;")
            db.query("install aws; load aws;")
            db.query(f"CALL load_aws_credentials('{self.aws_profile}');")
        db.query(self.options)
        return db

    def query(self, select_statement: SQL):
        db = self.connect()
        dataframes = select_statement.collect_dataframes()
        for key, value in dataframes.items():
            db.register(key, value)

        result = db.query(select_statement.to_string())
        if result is None:
            return
        return result.df()

    def __enter__(self):
        self.__connection = self.connect()
        return self.__connection

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.__connection.close()
        self.__connection = None
